==> experience_builder-0.1.0-alpha1.tar.gz: experience_builder/src/Plugin/DataType/ComponentTreeHydrated.php <==
<?php

declare(strict_types=1);

namespace Drupal\experience_builder\Plugin\DataType;

use Drupal\Core\Cache\CacheableDependencyInterface;
use Drupal\Core\Cache\CacheableJsonResponse;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Entity\Plugin\DataType\EntityAdapter;
use Drupal\Core\Render\RenderableInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TypedData\Attribute\DataType;
use Drupal\Core\TypedData\TypedData;
use Drupal\Core\Theme\ComponentPluginManager;
use Drupal\experience_builder\Entity\Component;
use Drupal\experience_builder\Plugin\Field\FieldType\ComponentTreeItem;

/**
 * @todo Do we need multiple variations of this? See \Drupal\datetime\DateTimeComputed for an example where there's *settings*
 * @phpstan-type JsonSchema array<string, mixed>
 */
#[DataType(
  id: "component_tree_hydrated",
  label: new TranslatableMarkup("Hydrated component tree"),
  description: new TranslatableMarkup("Computed from tree structure + props values"),
)]
class ComponentTreeHydrated extends TypedData implements CacheableDependencyInterface, RenderableInterface {

  /**
   * @param JsonSchema $slot_schema
   */
  private static function getDefaultSlotValue(array $slot_schema): string {
    if (array_key_exists('examples', $slot_schema) && array_key_exists(0, $slot_schema['examples'])) {
      return $slot_schema['examples'][0];
    }
    return '';
  }

  private function getComponentPluginManager(): ComponentPluginManager {
    return \Drupal::service(ComponentPluginManager::class);
  }

  /**
   * {@inheritdoc}
   */
  public function getValue(): CacheableJsonResponse {
    $item = $this->getParent();
    assert($item instanceof ComponentTreeItem);
    $tree = $item->get('tree');
    assert($tree instanceof ComponentTreeStructure);

    $hydrated = [];

    // Hydrate all component instances, but only considering props. This
    // essentially means getting the values for each component instance, while
    // ignoring their slots. The result: a flat list of hydrated components, but
    // with all slots empty.
    foreach ($tree->getComponentInstanceUuids() as $uuid) {
      $component_id = $tree->getComponentId($uuid);
      $sdc_component_props = $item->resolveComponentProps($uuid);

      // Use the first example defined in SDC metadata, if it exists. Otherwise,
      // fall back to `"#plain_text => ''`, which is accepted by SDC's rendering
      // logic but still results in an empty slot.
      // @see https://www.drupal.org/node/3391702
      // @see \Drupal\Core\Render\Element\ComponentElement::generateComponentTemplate()
      // @see \Drupal\experience_builder\Controller\SdcController::wrapComponentsForPreview()
      $sdc_component_plugin_instance = $this->getComponentPluginManager()->find(Component::convertIdToMachineName($component_id));
      $default_slot_values = array_map(
        fn (array $s): string => self::getDefaultSlotValue($s),
        $sdc_component_plugin_instance->metadata->slots,
      );

      $hydrated[$uuid] = [
        'component' => $component_id,
        'props' => $sdc_component_props,
      ];
      if (!empty($sdc_component_plugin_instance->metadata->slots)) {
        $hydrated[$uuid]['slots'] = $default_slot_values;
      }
    }

    // Transform the flat list of hydrated components into a hydrated component
    // tree, by assigning child components to their parent component's slot. If
    // this happens depth-first, then the tree will gradually be built, with the
    // last iteration assigning the last component to the component tree's root.
    foreach ($tree->getSlotChildrenDepthFirst() as $parent_uuid => ['slot' => $slot, 'uuid' => $uuid]) {
      assert(array_key_exists('slots', $hydrated[$parent_uuid]));

      // Remove default slot value: this slot is populated.
      if (is_string($hydrated[$parent_uuid]['slots'][$slot])) {
        unset($hydrated[$parent_uuid]['slots'][$slot]);
      }

      $hydrated[$parent_uuid]['slots'][$slot][$uuid] = $hydrated[$uuid];
      unset($hydrated[$uuid]);
    }

    return (new CacheableJsonResponse())
      ->addCacheableDependency($this->getCacheability())
      ->setData([ComponentTreeStructure::ROOT_UUID => $hydrated]);
  }

  /**
   * {@inheritdoc}
   */
  public function setValue($value, $notify = TRUE): void {
    // There is nothing to set, so return early.
    // @todo This is an upstream core bug, because this is defined as both computed and read-only, yet it still gets called 🙃 Once the core bug is fixed, this should throw a ReadOnlyException.
    // throw new ReadOnlyException();
  }

  /**
   * {@inheritdoc}
   */
  public function toRenderable(): array {
    // ⚠️ We *could* convert to a render array directly. But that should not be
    // the source of truth. So we start from a Drupal Render API-agnostic point,
    // and map that into a render array. This guarantees none of this will ever
    // rely on Render API specifics.
    $renderable_component_tree = $this->getValue();

    $build = [];
    // @see \Drupal\Core\Entity\EntityViewBuilder::getBuildDefaults()
    $renderable_component_tree->getCacheableMetadata()->applyTo($build);

    $json = $renderable_component_tree->getContent();
    assert(is_string($json));
    $hydrated = json_decode($json, TRUE);

    assert(array_keys($hydrated) === [ComponentTreeStructure::ROOT_UUID]);
    return self::renderify($hydrated);
  }

  /**
   * Recursively converts an array generated by ::getValue() to a render array.
   *
   * @param array $hydrated
   *   An array generated by ::getValue().
   *
   * @return array
   *   The corresponding render array.
   */
  private static function renderify(array $hydrated) {
    $build = [];
    foreach ($hydrated as $component_subtree_uuid => $component_instances) {
      foreach ($component_instances as $component_instance_uuid => $component_instance) {
        // @todo This will need to evolve when supporting non-SDC component types in https://www.drupal.org/project/experience_builder/issues/3454519
        $build[$component_subtree_uuid][$component_instance_uuid] = [
          '#type' => 'component',
        ];
        foreach ($component_instance as $key => $value) {
          if ($key !== 'slots') {
            // An SDC plugin ID is expected when using `#type => component`.
            // @see \Drupal\Core\Render\Element\ComponentElement
            if ($key === 'component') {
              $value = Component::convertIdToMachineName($value);
            }

            // Note: this works because `::getValue()` above uses `props` and
            // `slots`, which allow simply prefixing them with `#` to generate the
            // corresponding render array.
            $build[$component_subtree_uuid][$component_instance_uuid]["#$key"] = $value;
            continue;
          }

          $build[$component_subtree_uuid][$component_instance_uuid]["#slots"] = [];
          foreach ($value as $slot => $slot_value) {
            // Handle default slot value: convert to renderable using either
            // #plain_text or `#markup`.
            if (is_string($slot_value)) {
              $build[$component_subtree_uuid][$component_instance_uuid]["#slots"][$slot] = !str_starts_with($slot_value, '<')
                // Match how Drupal core handles string values for components.
                // @see https://www.drupal.org/node/3398039
                ? ['#plain_text' => $slot_value]
                // TRICKY: this goes beyond what Drupal core appears to allow
                // in https://www.drupal.org/node/3398039, but the SDC plugin
                // manager accepts this as a valid SDC definition, so XB has
                // no choice but to support it.
                : ['#markup' => $slot_value];
            }
            // Explicit slot value: renderify, just like the rest of the
            // component tree.
            else {
              $build[$component_subtree_uuid][$component_instance_uuid]["#slots"] += self::renderify([$slot => $slot_value]);
            }
          }
        }
      }
    }
    return $build;
  }

  /**
   * Computes the cacheability of this computed property.
   *
   * @return \Drupal\Core\Cache\CacheableMetadata
   *   The cacheability of the computed value.
   */
  private function getCacheability(): CacheableMetadata {
    // @todo Once bundle-level defaults for `tree` + `props` are supported, this should also include cacheability of whatever config that is stored in.
    // @see \Drupal\experience_builder\Plugin\Field\FieldType\ComponentTreeItem::preSave()

    $root = $this->getRoot();
    if ($root instanceof EntityAdapter) {
      return CacheableMetadata::createFromObject($root->getEntity());
    }

    // This appears to be an ephemeral component tree, hence it is uncacheable.
    return (new CacheableMetadata())->setCacheMaxAge(0);
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTags() {
    return $this->getCacheability()->getCacheTags();
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheContexts() {
    return $this->getCacheability()->getCacheContexts();
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheMaxAge() {
    return $this->getCacheability()->getCacheMaxAge();
  }

}

==> experience_builder-0.2.1-alpha1.tar.gz: experience_builder/src/Plugin/DataType/ComponentTreeHydrated.php <==
<?php

declare(strict_types=1);

namespace Drupal\experience_builder\Plugin\DataType;

use Drupal\Core\Cache\CacheableDependencyInterface;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Entity\Plugin\DataType\EntityAdapter;
use Drupal\Core\Render\Markup;
use Drupal\Core\Render\RenderableInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TypedData\Attribute\DataType;
use Drupal\Core\TypedData\TypedData;
use Drupal\experience_builder\ComponentSource\ComponentSourceWithSlotsInterface;
use Drupal\experience_builder\Entity\Component;
use Drupal\experience_builder\HydratedTree;
use Drupal\experience_builder\Plugin\Field\FieldType\ComponentTreeItem;

/**
 * @todo Do we need multiple variations of this? See \Drupal\datetime\DateTimeComputed for an example where there's *settings*
 */
#[DataType(
  id: "component_tree_hydrated",
  label: new TranslatableMarkup("Hydrated component tree"),
  description: new TranslatableMarkup("Computed from tree structure + input values"),
)]
class ComponentTreeHydrated extends TypedData implements CacheableDependencyInterface, RenderableInterface {

  /**
   * {@inheritdoc}
   */
  public function getValue(): array {
    $item = $this->getParent();
    assert($item instanceof ComponentTreeItem);
    $tree = $item->get('tree');

    $hydrated = [];

    // Load all unique Components used in this tree in a single query.
    $components = Component::loadMultiple($tree->getComponentIdList());

    // Hydrate all component instances, but only considering props. This
    // essentially means getting the values for each component instance, while
    // ignoring their slots. The result: a flat list of hydrated components, but
    // with all slots empty.
    foreach ($tree->getComponentInstanceUuids() as $uuid) {
      $component_id = $tree->getComponentId($uuid);
      $component = $components[$component_id];
      assert($component instanceof Component);

      $source = $component->getComponentSource();
      $hydrated[$uuid] = [
        'component' => $component_id,
      ] + $source->hydrateComponent($source->getExplicitInput($uuid, $item));
    }

    // Transform the flat list of hydrated components into a hydrated component
    // tree, by assigning child components to their parent component's slot. If
    // this happens depth-first, then the tree will gradually be built, with the
    // last iteration assigning the last component to the component tree's root.
    foreach ($tree->getSlotChildrenDepthFirst() as $parent_uuid => ['slot' => $slot, 'uuid' => $uuid]) {
      assert(array_key_exists('slots', $hydrated[$parent_uuid]));

      // Remove default slot value: this slot is populated.
      if (is_string($hydrated[$parent_uuid]['slots'][$slot])) {
        unset($hydrated[$parent_uuid]['slots'][$slot]);
      }

      $hydrated[$parent_uuid]['slots'][$slot][$uuid] = $hydrated[$uuid];
      unset($hydrated[$uuid]);
    }
    return [ComponentTreeStructure::ROOT_UUID => $hydrated];
  }

  /**
   * {@inheritdoc}
   */
  public function setValue($value, $notify = TRUE): void {
    // There is nothing to set, so return early.
    // @todo This is an upstream core bug, because this is defined as both computed and read-only, yet it still gets called 🙃 Once the core bug is fixed, this should throw a ReadOnlyException.
    // throw new ReadOnlyException();
  }

  /**
   * {@inheritdoc}
   */
  public function toRenderable(bool $isPreview = FALSE): array {
    // ⚠️ We *could* convert to a render array directly. But that should not be
    // the source of truth. So we start from a Drupal Render API-agnostic point,
    // and map that into a render array. This guarantees none of this will ever
    // rely on Render API specifics.
    $renderable_component_tree = $this->getHydratedTree();

    $build = [];
    // @see \Drupal\Core\Entity\EntityViewBuilder::getBuildDefaults()
    CacheableMetadata::createFromObject($renderable_component_tree)->applyTo($build);

    $hydrated = $renderable_component_tree->getTree();

    assert(array_keys($hydrated) === [ComponentTreeStructure::ROOT_UUID]);
    return self::renderify($hydrated, $isPreview);
  }

  /**
   * Recursively converts an array generated by ::getValue() to a render array.
   *
   * @param array $hydrated
   *   An array generated by ::getValue().
   * @param bool $isPreview
   *   TRUE if is preview.
   *
   * @return array
   *   The corresponding render array.
   */
  private static function renderify(array $hydrated, bool $isPreview = FALSE) {
    $build = [];
    foreach ($hydrated as $component_subtree_uuid => $component_instances) {
      foreach ($component_instances as $component_instance_uuid => $component_instance) {
        $component = Component::load($component_instance['component']);
        assert($component instanceof Component);
        $source = $component->getComponentSource();
        $build[$component_subtree_uuid][$component_instance_uuid] = $source->renderComponent($component_instance, $component_instance_uuid, $isPreview);

        // Wrap each rendered component instance in HTML comments that allow the
        // client side to identify it.
        $build[$component_subtree_uuid][$component_instance_uuid]['#prefix'] = Markup::create("<!-- xb-start-$component_instance_uuid -->");
        $build[$component_subtree_uuid][$component_instance_uuid]['#suffix'] = Markup::create("<!-- xb-end-$component_instance_uuid -->");

        // Associate the `Component` config entity cache tag with every rendered
        // component instance — remove the need for each `ComponentSource`
        // plugin to do this in an awkward way in their `::renderComponent()`.
        // This also associates any cache contexts and max-age; both may be used
        // for dynamic config overrides.
        CacheableMetadata::createFromRenderArray($build[$component_subtree_uuid][$component_instance_uuid])
          ->addCacheableDependency($component)
          ->applyTo($build[$component_subtree_uuid][$component_instance_uuid]);

        // Figure out the slots, if there are any.
        if ($source instanceof ComponentSourceWithSlotsInterface  && !empty($component_instance['slots'])) {
          $slots = [];
          foreach ($component_instance['slots'] as $slot => $slot_value) {
            // Handle default slot value: convert to renderable using either
            // #plain_text or `#markup`.
            if (is_string($slot_value)) {
              $slots[$slot] = !str_starts_with($slot_value, '<')
                // Match how Drupal core handles string values for components.
                // @see https://www.drupal.org/node/3398039
                ? ['#plain_text' => $slot_value]
                // TRICKY: this goes beyond what Drupal core appears to allow
                // in https://www.drupal.org/node/3398039, but the SDC plugin
                // manager accepts this as a valid SDC definition, so XB has
                // no choice but to support it.
                : ['#markup' => $slot_value];
            }
            // Explicit slot value: renderify, just like the rest of the
            // component tree.
            else {
              $slots += self::renderify([$slot => $slot_value]);
            }
          }

          $source->setSlots($build[$component_subtree_uuid][$component_instance_uuid], $slots);
        }
      }
    }
    return $build;
  }

  /**
   * Computes the cacheability of this computed property.
   *
   * @return \Drupal\Core\Cache\CacheableMetadata
   *   The cacheability of the computed value.
   */
  private function getCacheability(): CacheableMetadata {
    // @todo Once bundle-level defaults for `tree` + `inputs` are supported, this should also include cacheability of whatever config that is stored in.
    // @see \Drupal\experience_builder\Plugin\Field\FieldType\ComponentTreeItem::preSave()

    $root = $this->getRoot();
    if ($root instanceof EntityAdapter) {
      return CacheableMetadata::createFromObject($root->getEntity());
    }

    // This appears to be an ephemeral component tree, hence it is uncacheable.
    return (new CacheableMetadata())->setCacheMaxAge(0);
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTags() {
    return $this->getCacheability()->getCacheTags();
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheContexts() {
    return $this->getCacheability()->getCacheContexts();
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheMaxAge() {
    return $this->getCacheability()->getCacheMaxAge();
  }

  /**
   * Gets the tree as a hydrated tree value object.
   */
  public function getHydratedTree(): HydratedTree {
    return new HydratedTree(
      $this->getValue(),
      $this->getCacheability(),
    );
  }

}
